import requests
import logging
import json
from . import v3
from . import v4
from .base import BaseApi
from .errors import ApiError
from .errors import RequestError


logger = logging.getLogger('civicrmapi')


class BaseRestApi(BaseApi):
    def __init__(self, url, htaccess=None, verify_ssl=True, timeout=None, headers=None):
        super().__init__()
        self.url = url
        self.verify_ssl = verify_ssl
        self.timeout = timeout
        self.headers = headers

        # Setup basic-auth
        if htaccess:
            auth_user = htaccess['user']
            auth_pass = htaccess['pass']
            self.auth = requests.auth.HTTPBasicAuth(auth_user, auth_pass)
        else:
            self.auth = None

    def _perform_post_request(self, params, url_path=None):
        if url_path:
            url = '{}/{}'.format(self.url.rstrip('/'), url_path.lstrip('/'))
        else:
            url = self.url
        logger.info(f'Perform post request: {url}')
        logger.debug(f'- params: {params}')
        logger.debug(f'- headers: {self.headers}')
        logger.debug(f'- verify: {self.verify_ssl}')
        logger.debug(f'- timeout: {self.timeout}')

        try:
            reply = requests.post(
                url,
                params=params,
                verify=self.verify_ssl,
                auth=self.auth,
                timeout=self.timeout,
                headers=self.headers)
        except requests.exceptions.RequestException as exc:
            raise RequestError(exc)
        else:
            logger.info(f'Post request done: {reply}')
            logger.debug(f'- text: {reply.text}')

        if not reply.status_code == 200:
            raise ApiError(reply)
        else:
            return reply.text

    def _perform_api_call(self, entity, action, params):
        raise NotImplemented


class RestApiV3(BaseRestApi):
    """
    Simple bindings for CiviCRM's RestApiv3.
    """
    VERSION = v3

    def __init__(self, url, api_key, site_key, htaccess=None, verify_ssl=True, timeout=None):
        """
        :param str url: CiviCRM's base-url
        :param str api_key: CiviCRM's api-key
        :param str site_key: CiviCRM's api-site_key
        :param dict htaccess: htaccess credentials with 'user' and 'pass' keys. (optional)
        :param bool verify_ssl: Verify SSL-certificate or not. Default is True. (optional)
        :param int timeout: Timeout in seconds. (optional)
        """
        self.api_key = api_key
        self.site_key = site_key
        url = url.rstrip('/') + '/civicrm/ajax/rest'
        super().__init__(url, htaccess, verify_ssl, timeout)

    def _perform_api_call(self, entity, action, params):
        params['sequential'] = params.get('sequential', 1)
        base_params = dict()
        base_params['api_key'] = self.api_key
        base_params['key'] = self.site_key
        base_params['entity'] = entity
        base_params['action'] = action
        base_params['json'] = json.dumps(params)
        return self._perform_post_request(base_params)


class RestApiV4(BaseRestApi):
    """
    Simple bindings for CiviCRM's RestApiv4.
    """
    VERSION = v4

    def __init__(self, url, api_key, htaccess=None, verify_ssl=True, timeout=None):
        """
        :param str url: CiviCRM's base-url
        :param str api_key: CiviCRM's api-key
        :param dict htaccess: htaccess credentials with 'user' and 'pass' keys. (optional)
        :param bool verify_ssl: Verify SSL-certificate or not. Default is True. (optional)
        :param int timeout: Timeout in seconds. (optional)
        """
        self.api_key = api_key
        url = url.rstrip('/') + '/civicrm/ajax/api4/'
        headers = dict()
        headers['Content-Type'] = 'application/x-www-form-urlencoded'
        headers['X-Civi-Auth'] = 'Bearer {}'.format(self.api_key)
        super().__init__(url, htaccess, verify_ssl, timeout, headers)

    def _perform_api_call(self, entity, action, params):
        params = dict(params=json.dumps(params))
        url_path = f'{entity}/{action}'
        return self._perform_post_request(params, url_path)


import invoke
import shlex
import json
import logging
from . import v3
from . import v4
from .base import BaseApi
from .errors import InvokeError
from .errors import ApiError


logger = logging.getLogger('civicrmapi')


class BaseConsoleApi(BaseApi):
    def __init__(self, cv, cwd=None, context=None):
        """
        :param str cv: cv command
        :param str cwd: working directory for cv
        :param str context:
            If a context is given the original command will be tokenized and
            given to the context as positional argument. The simplest context
            migth be a 'sh -c'. But it could also be a docker-exec- or
            ssh-command to run the api call within a docker container or on a
            remote machine.
        """
        super().__init__()
        self.cv = shlex.split(cv)
        self.cwd = ['--cwd', shlex.quote(f'{cwd}')] if cwd else list()
        self.context = context

    def _run(self, command):
        if self.context:
            context = shlex.split(self.context)
            command = '{} {}'.format(' '.join(context), shlex.quote(command))
        logger.info(f'Run command: {command}')
        try:
            reply = invoke.run(command, hide=True)
        except invoke.exceptions.UnexpectedExit as exc:
            raise ApiError(exc.result)
        except invoke.exceptions.Failure as exc:
            raise InvokeError(exc)
        else:
            logger.info(f'Running command finished [{reply.return_code}]')
            logger.debug(f'- stdout: {reply.stdout}')
            logger.debug(f'- stderr: {reply.stderr}')

        if reply.stderr and not reply.stdout:
            raise ApiError(reply)
        else:
            return reply

    def _get_command(self, entity, action, params):
        raise NotImplemented

    def _perform_api_call(self, entity, action, params):
        command = self._get_command(entity, action, params)
        reply = self._run(command)
        return reply.stdout


class ConsoleApiV3(BaseConsoleApi):
    VERSION = v3

    def _get_command(self, entity, action, params):
        params['sequential'] = 1
        echo_params = ['echo', shlex.quote(json.dumps(params))]
        api = ['api3', shlex.quote(f'{entity}.{action}'), '--in=json']
        command = self.cv + self.cwd + api
        return '{} | {}'.format(' '.join(echo_params), ' '.join(command))


class ConsoleApiV4(BaseConsoleApi):
    VERSION = v4

    def _get_command(self, entity, action, params):
        api = ['api4', shlex.quote(f'{entity}.{action}')]
        params = [shlex.quote(json.dumps(params))] if params else list()
        command = self.cv + self.cwd + api + params
        return ' '.join(command)




import logging
import json
from .errors import ApiError
from .errors import InvalidJSON


logger = logging.getLogger('civicrmapi')


class BaseAction:
    ENTITY = None
    ACTION = None

    def __init__(self, api):
        if not self.ENTITY or not self.ACTION:
            raise NotImplemented('ENTITY and ACTION must be defined.')
        self._api = api

    def __call__(self, params=None):
        """
        :param dict params: api call parameters (optional)
        :return dict: api call result
        """
        return self._api(self.ENTITY, self.ACTION, params or dict())


class BaseEntity:
    ENTITY = None
    ACTIONS = list()

    def __init__(self, api):
        if not self.ENTITY:
            raise NotImplemented('ENTITY must be defined.')
        self._api = api
        self._add_actions()

    def _add_actions(self):
        for action in self.ACTIONS + self._api.VERSION.ACTIONS:
            if isinstance(action, str):
                action_name = action
                attrs = dict(ENTITY=self.ENTITY, ACTION=action)
                action_class = type(action, (BaseAction,), attrs)
            elif isinstance(action, type) and issubclass(action, BaseAction):
                action_name = action.__name__
                action_class = action
            else:
                msg = 'ACTIONS item must be string or subclass of BaseAction.'
                raise ValueError(msg)
            if not hasattr(self, action_name):
                setattr(self, action_name, action_class(self._api))

    def __call__(self, action, params=None):
        """
        :param str action: api call action
        :param dict params: api call parameters (optional)
        :return dict: api call result
        """
        return self._api(self.ENTITY, action, params or dict())


class BaseApi:
    VERSION = None
    ENTITIES = list()

    def __init__(self):
        if not self.VERSION:
            raise NotImplemented('VERSION must be defined.')
        self._add_entities()

    def _add_entities(self):
        for entity in self.ENTITIES + self.VERSION.ENTITIES:
            if isinstance(entity, str):
                entity_name = entity
                entity_class = type(entity, (BaseEntity,), dict(ENTITY=entity))
            elif isinstance(entity, type) and issubclass(entity, BaseEntity):
                entity_name = entity.__name__
                entity_class = entity
            else:
                msg = 'ENTITIES item must be string or subclass of BaseEntity.'
                raise ValueError(msg)
            if not hasattr(self, entity_name):
                setattr(self, entity_name, entity_class(self))

    def __call__(self, entity, action, params=None):
        """
        :param str entity: CiviCRM-entitiy
        :param str action: api call action
        :param dict params: api call parameters (optional)
        :return dict: normalized api call result
        :raises RequestError: when the rest api could not be accessed
        :raises InvokeError: when the console api could not be accessed
        :raises ApiError: when the api call failed
        :raises InvalidJson: when the response is invalid json code
        """
        logger.info(f'Perform api call: {entity}.{action} with {params}')
        result = self._perform_api_call(entity, action, params or dict())
        return self._process_json_result(result)

    def _perform_api_call(self, entity, action, params):
        raise NotImplemented

    def _process_json_result(self, json_response):
        """
        Called by __call__ to process json response.
        """
        try:
            result = json.loads(json_response)
        except json.JSONDecodeError as exc:
            raise InvalidJSON(exc)
        else:
            logger.info(f'Valid json response.')
            logger.debug(f'Decoded json: {result}')

        # Check api-v3 result for an api error.
        if isinstance(result, dict) and result.get('is_error', False):
            raise ApiError(result)
        # Check api-v4 result for an api error.
        elif isinstance(result, dict) and 'error_code' in result:
            raise ApiError(result)
        else:
            return self._normalize_result_values(result)

    def _normalize_result_values(self, result):
        # Console api v4 returns values as list.
        if isinstance(result, list):
            return result
        # If we have a dict with values we return only those.
        elif 'values' in result:
            return result['values']
        # Returning the result as dict as a fallback.
        # (That should not happen at all at all.)
        else:
            return result



from .base import BaseAction
from .base import BaseEntity
from .base import BaseApi
from .errors import BaseException
from .errors import RequestError
from .errors import InvokeError
from .errors import InvalidJSON
from .errors import ApiError
from .rest import RestApiV3
from .rest import RestApiV4
from .console import ConsoleApiV3
from .console import ConsoleApiV4


import json
import requests
import invoke


class BaseException(Exception):
    """
    All exceptions possibly raised by civicrmapi inherit from BaseException.
    """
    pass

class RequestError(BaseException):
    """
    A simple wrapper for all exceptions raised by requests.
    """
    pass


class InvokeError(BaseException):
    """
    A simple wrapper for all exceptions raised by invoke.
    """
    pass


class InvalidJSON(BaseException):
    """
    A simple wrapper for a json.JSONDecodeError.
    """
    pass


class ApiError(BaseException):
    """
    Raised for all api related errors as well as return-code and http errors.
    """
    def __init__(self,value, msg=None):
        self.value = value
        self.msg = msg

    def _as_json(self, value):
        try:
            result = json.loads(value)
        except json.JSONDecodeError:
            return value
        else:
            return json.dumps(result, sort_keys=True, indent=4)

    def __str__(self):
        if self.msg:
            msg = [f'ERROR: {self.msg}']
        else:
            msg = list()
        if isinstance(self.value, requests.Response):
            msg.append(f'URL: {self.value.url}')
            msg.append(f'HTTP-CODE: {self.value.status_code}')
            msg.append('RESPONSE: {}'.format(self._as_json(self.value.text)))
        elif isinstance(self.value, invoke.runners.Result):
            msg.append(f'COMMAND: {self.value.command}')
            msg.append(f'RETURN-CODE: {self.value.return_code}')
            msg.append('STDOUT: {}'.format(self._as_json(self.value.stdout)))
            msg.append('STDERR: {}'.format(self._as_json(self.value.stderr)))
        elif isinstance(self.value, dict) or isinstance(self.value, list):
            msg.append(json.dumps(self.value, sort_keys=True, indent=4))
        else:
            msg.append(self.value)
        return '\n'.join(msg)
